#
# Backup script as detailed in GAG-News 173 (see www.gag.de)
#
# Copyright 2020, Herbert zur Nedden
# Permission to use and amend
#
# This script is supplied as-is and the useage is at YOUR OWN RISK!
# I did some testing and for me it does work - but that is no promise or guarantee it will work for you!
#
# Global concept is that all things inside $MyOriginal are to be backed up.
# The files and folders (with subfolders) matching $MyEye are backed up to the cloud, the rest is backed up locally
# The files are put into encrypted 7zip archives for the backup into $MyCloud and $MyLocal depending on where they shall go
# From $MyCloud and $MyLocal things are then transferred to $ToCloud, $ToUSB; and from $MyOriginal to $ToNAS
# Backup is thus in these stages
# 1. Fetch things like Outlook files and insert them into $MyOriginal
# 2. Split $MyOriginal into two folders $MyCloud and $MyLocal for later backup to the cloud and USB drive
# 3. Synchronize $MyCloud and $MyLocal with the cloud storage and/or local backup on USB drive; $MyOriginal with NAS
# Sorry, for any further detail you need to learn from the code below, or read GAG-News :-)

# In addition there is some code to fetch my .ost and .pst Outlook stores from the system disc into $MyOrginal

Param([string]$ParmCommand = "", [string]$ParmMask="*")


if (!(Test-Path ".\Communary.FileExtensions\Communary.FileExtensions.psm1")) {
	Write-Host ""
	Write-Host "Something is missing:"
	Write-Host ""
	Write-Host "You need the Communary.FileExtensions that you can find here: https://github.com/gravejester/Communary.FileExtensions"
	Write-Host "Just fetch https://github.com/gravejester/Communary.FileExtensions/archive/master.zip from that pages"
	Write-Host "Then unpack it into the folder Communary.FileExtensions next to this script"
	Write-Host ""
	exit
}

Import-Module .\Communary.FileExtensions\Communary.FileExtensions.psm1
# Invoke-FastFind -Path xxx -File -Recurse -System -Hidden -AttributeFilterMode Exclude (I do not want hidden or system files!)
# Output contains these Name, Path, FileSize, LastWriteTime (and some more)
# The reason I use these is due to how Get-ChildItem works (details are in GAG-News :-))

$MyOdd   = "Y" # Cloud folder for non A-Z file names

$MyCharsLen = $MyChars.length

# The source files to be backed up
$MyOriginal = "D:\MyData"                  # This is the source to be backed up
$MyIgnore   = "D:\MyData\Backup"           # This path part is ignored when defining the cloud backup folder name
$MySkip     = "\~"                         # Mask for file and folder names to not back up (the office temp stuff)
$MyAsIs     = "C:\Users\...\Keepass.kdbx"  # Add this one unencrypted																	 

# Some other files I add into the source files to be backed up
$MyOST      = "C:\Users\...\AppData\Local\Microsoft\Outlook\*.ost"
$MyPST      = "C:\Users\...\Documents\Outlook-Dateien\*.pst"
$MyOutlook  = $MyOriginal + "\Backup\Outlook"

# The backup set I fill and the folder into which I retrieve
$MyCloud    = "D:\MyBackup\Cloud"          # This is the subfolder for the things to go to the cloud
$MyLocal    = "D:\MyBackup\Local"          # This is the subfolder for the things to go to local backups
$MyRetrieve = "D:\MyBackup\Restore"        # Into this folder things are restored from backup
$MyGone     = "D:\MyBackup\Gone"           # Into this folder things deleted from the source are moved just in case..
$MyEye      = "*_\*"                       # Mask for file and folder names of things to be put into the cloud

# Cloud Credentials (used to connect to cloud storage with WebDAV)
$MyCloudPath = "https://..."
$MyCloudUser = "..."
$MyCloudPass = "..."

# This is where the backup will then be transferred to
$ToCloud    = "E:\MyCloud"                 # This it the pointer to the cloud backup folder
$ToUSB      = "F:\MyUSB"                   # This it the pointer to the local USB drive backup
$ToNAS      = "G:\MyNAS"                   # This it the pointer to the local NAS backup

# The log file goes to the TEMP folder
$MyLog      = $Env:TEMP + "\EnDeCode*.log" # My log file (the * will be replaced by the command name)

# MKDir
function mkdir([String] $What) {
	if (!(Test-Path $What)) { New-Item -ItemType directory -Path $What | Out-Null }
}
# MKPath (i.e. MKDir for the folder in the given path)
function mkpath([String] $What) {
	mkdir (Split-Path $What)
}

# Log a bit - second parameter controls screen output
function Blabla([string]$Bla, [string]$Con) {
	$Bla | Out-File -FilePath $MyLog -Encoding UTF8 -Append
	switch ($Con) {
	"."    { write-host "." -NoNewLine; }         # Just a progress dot
	".*"   { write-host " "; write-host $Bla; }   # Text but with a LF since progress dotes came before
	"*"    { write-host $Bla; }                   # Text as is
	}
}

# Convert file name into something cryptic
function Encode([String] $What) {
	# No chars that are part of hex strings in $GoodChars!
	$GoodChars = "GHIJKLMNOPQRSTUVWXYZghijklmnopqrstuvwxyz-_."
	# Get the folder name from the first letter of the source file
	$x = (Split-Path $What -File).substring($i,1).ToUpper()
	# Ensure a good character or put an unserscore
	if (!($GoodChars.contains($x))) { $x = "_"; }
	$Encoded = $x + "\";
	for ($i = 0; $i -lt $What.length; $i++) {
		$c = $What.substring($i,1);
		if ($GoodChars.contains($c)) {
			$Encoded += $c
		}
		else {
			$Encoded += $([System.Convert]::ToString([int][char]$c,16))
		}
	}
	$Encoded
}

# Revert from cryptic file name back to the real thing
function Decode([String] $What) {
	$HexChars = "abcdef0123456789"
	$Decoded = ""
	for ($i = 0; $i -lt $What.length; $i++) {
		$c = $What.substring($i,1);
		if ($HexChars.contains($c)) {
			$i++; $c += $What.substring($i,1);
			$Decoded += [char]$([System.Convert]::ToInt32($c,16))
		}
		else {
			$Decoded += $c;
		}
	}
	$Decoded
}

# Create some password
Function Garble([String] $What) {
	if ($What -eq "") { exit 0 }
	# Use the file name without path
	if ($What.contains("\")) { $Work = Split-Path $What -Leaf; }
	else                     { $Work = $What;                  }
	# Ensure we have at least 36 chars since I need 3x 12 chars
	do {
		$Work += "!"
		$Work += $Work
	} while ($Work.length -lt 36)
	$Garbled = $Work.substring(0,12) + "H@h@" + $work.substring(12,12) + "H0h0" + $work.substring(24,12) + "?"
	$Garbled
}


# Copy from original folder to backup folder (splitting up for cloud and local)
function BackupOne([string] $What) {
	# Ensure that a few second time shift will not change anything but the second of the time stamp
	TimeAdjust $What
	$f = Get-ChildItem -Path $What
	$srcPath = $f.VersionInfo.FileName
	if ($srcPath -like $MyEye) { 
		# This one goes to the cloud and thus gets a anonymized file name
		# The cloud path is the first letter of the source file name
		# Reason for this to try to not have too many files in one of the cloud folders
		$srcName = Split-Path $srcPath -Leaf
		$dstPath = $MyCloud +"\"+ $(Encode($srcPath)) 
	}
	else {
		# This one is backed up locally
		$dstPath = $srcPath.replace($MyOriginal,$MyLocal) 
	}
	mkpath $dstPath
	$wrkFolder = "-w" + (Split-Path $dstPath)
	$dstPass = "-p$(Garble($srcPath))"
	#Command "C:\Program Files\7-Zip\7z.exe"
	# a					Add to archive
	# Archivname
	# -sae				No archive name extension
	# -aoa				Overwrite brutally
	# -mx0				Copy only
	# -mhc=on			Archive header compessing
	# -mhe=on			Archive header encryption
	# -scrcsha256 		SHA256 Hash
	# -p"..."			Password
	# -slp				Largs pages for speed
	# -t7z				Type of archive
	# -w 				Work folder
	# -y				Do not ask nothing
	# Sourcefile
	if (Test-Path $dstPath) { Remove-Item $dstPath }
	if ($What -eq $MyAsIs) {
		Copy-Item $srcPath $dstPath -Force
	}
	else {
		& 'C:\Program Files\7-Zip\7z.exe' a $dstPath -sae -aoa -mx0 -mhc=on -mhe=on -scrcsha256 -slp -t7z $wrkFolder -y $dstPass $srcPath | Out-Null
		(Get-ChildItem $dstPath).CreationTime   = $f.CreationTime
		(Get-ChildItem $dstPath).LastWriteTime  = $f.LastWriteTime
		(Get-ChildItem $dstPath).LastAccessTime = $f.LastAccessTime
	}
	#Write-Host "Encrypted" $srcPath "->" $dstPath
	Blabla "Encrypted`t$($srcPath)`t$($dstPath)"
}

# Adjust timestamp to avoid overflow or underflow
function TimeAdjust([string] $What) {

	$f = Get-ChildItem -Path $What

	if ($f.LastWriteTime.Second -lt 3) {
		(Get-ChildItem $f).LastWriteTime = $f.LastWriteTime.AddSeconds(3)
	}
	if ($f.LastWriteTime.Second -gt 57) {
		(Get-ChildItem $f).LastWriteTime = $f.LastWriteTime.AddSeconds(-3)
	}
}

# Copy from backup (be it cloud or local) folder to retrieve folder
function Restore([string] $Mask) {
	$c = 0
	Blabla "Restoring $($Mask)" "*"
	Blabla "Scanning backup files from $($MyCloud) and $($MyLocal)"
	$SourceFiles = (Invoke-FastFind -Path $MyCloud -Recurse -File -System -Hidden -AttributeFilterMode Exclude) + (Invoke-FastFind -Path $MyLocal -Recurse -File -System -Hidden -AttributeFilterMode Exclude)
	Blabla "... I found $($SourceFiles.count) files"
	
	foreach ($f in $SourceFiles) {
		$srcName = $f.Name
		$srcPath = $f.Path
		if ($srcPath.contains($MyLocal)) { 
			# For locally backed up files set the destination path
			$dstPath = $srcPath.replace($MyLocal+"\","")
		}
		else { 
			# For things from cloud deanonymize the path
			$dstPath = Decode($srcName)
		}
		if ($dstPath -like $Mask) {
			$srcPass = "-p$(Garble($dstPath))"
			$dstPath = $MyRetrieve +"\" + $dstPath
			$dstFolder = "-o" + (Split-Path $dstPath) 
			mkpath $dstPath
			$wrkFolder = "-w" + (Split-Path $dstPath)
			#Command "C:\Program Files\7-Zip\7z.exe"
			# e					Extract from archive
			# Archivname
			# -p"..."			Password
			# -t7z				Type of archive
			# -w 				Work folder
			# -y				Do not ask nothing
			# -o...				Folder
			# Destinationfolder
			if (Test-Path $dstPath) { Remove-Item $dstPath }
			& 'C:\Program Files\7-Zip\7z.exe' e $srcPath -t7z $wrkFolder -y $srcPass $dstFolder | Out-Null
			Blabla "Decrypted`t$($srcPath)`t$($dstPath)" "."
			$c = $c+1
		}
	}
	Blabla "Restored $($Mask) - $($c) files" ".*"

}

# List from backup (scanning cloud or local) folder
function ListBackup([string] $Mask) {
	Blabla "Listing from Backup $($Mask)" "*"
	$ListFiles = (Invoke-FastFind -Path $MyCloud -Recurse -File -System -Hidden -AttributeFilterMode Exclude) + (Invoke-FastFind -Path $MyLocal -Recurse -File -System -Hidden -AttributeFilterMode Exclude)
	foreach ($f in $ListFiles) {
		$srcName = $f.Name
		$srcPath = $f.Path
		if ($srcPath.contains($MyLocal)) { 
			# For locally backed up files set the destination path
			$dstPath = $srcPath.replace($MyLocal+"\","")
			$srcName = "" 
		}
		else {
			# For things from cloud deanonymize the path but note the anonymized name
			$dstPath = Decode($srcName)
			$srcName = "["+$f.Name+"]" 
		}
		if ($dstPath -like $Mask) {
			Blabla "$dstPath $($srcName)" "*"
		}
	}
}

# List from original folder
function ListOriginal([string] $Mask) {
	Blabla "Listing from Original $($Mask)" "*"
	$ListFiles = Invoke-FastFind -Path $MyOriginal -Recurse -File -System -Hidden -AttributeFilterMode Exclude
	foreach ($f in $ListFiles) {
		$srcPath = $f.Path
		# Just cut off the base backup path
		$dstPath = $srcPath.replace($MyOriginal+"\","")
		if ($dstPath -like $Mask) {
			Blabla "$($dstPath)" "*"
		}
	}
}

# Compare original with backup (cloud or local) folder and update as needed
function Backup() {
	$cc = 0; $ca = 0; $cd = 0;
	Blabla "Backupping" "*"

	mkdir $MyLocal
	mkdir $MyCloud

	# Get the files and add MyFlag and for the cloud ones the decoded name field
	Blabla "Scanning original from $($MyOriginal)"
	$OriginalFiles = Invoke-FastFind -Path $MyOriginal -Recurse -File -System -Hidden -AttributeFilterMode Exclude | Select Name,LastWriteTime,Path,MyFlag,MyPath
	Blabla "Scanning backup's cloud part from $($MyCloud)"
	$CloudFiles    = Invoke-FastFind -Path $MyCloud    -Recurse -File -System -Hidden -AttributeFilterMode Exclude | Select Name,LastWriteTime,Path,MyFlag,MyPath
	Blabla "Scanning backup's local part from $($MyLocal)"
	$LocalFiles    = Invoke-FastFind -Path $MyLocal    -Recurse -File -System -Hidden -AttributeFilterMode Exclude | Select Name,LastWriteTime,Path,MyFlag,MyPath
	Blabla "... I found $($OriginalFiles.count) original files"
	Blabla "... I found $($CloudFiles.count) + $($LocalFiles.count) = $($CloudFiles.count + $LocalFiles.count) backup files"
	foreach ($sf in $OriginalFiles) {
		if ($($sf.Path).contains($MySkip)) {
			$sf.MyPath = "*skip*"
			$sf.MyFlag  = "skip"
		}
		else {
			$sf.MyPath = $($sf.Path).replace($MyOriginal+"\","")
			$sf.MyFlag  = "?"
		}
	}
	foreach ($df in $CloudFiles) {
		$df.MyPath = Decode($df.Name)
		$df.MyFlag  = "?"
	}
	foreach ($df in $LocalFiles) {
		$df.MyPath = $($df.Path).replace($MyLocal+"\","")
		$df.MyFlag  = "?"
	}

	$FilesSame = @()
	$FilesChanged = @()
	$FilesNew = @()
	$FilesOld = @()
	
	# Get the like named ones
	foreach ($sf in $OriginalFiles) {
		foreach ($df in $CloudFiles + $LocalFiles) {
			if ($sf.MyPath -eq $df.MyPath) {
				# Just time stamp since encrypting can change the file size...
				if ($sf.LastWriteTime -eq $df.LastWriteTime) {
					$sf.MyFlag = "same"
					$df.MyFlag = "same"
					$FilesSame += $sf.Path
					#Blabla "Same`t$($sf.MyPath)`t$($sf.LastWriteTime)`t$($df.LastWriteTime)"
				}
				else {
					$sf.MyFlag = "changed"
					$df.MyFlag = "changed"
					$FilesChanged += $sf.Path
					Blabla "Changed`t$($sf.MyPath)`t$($sf.LastWriteTime)`t$($df.LastWriteTime)"
				}
			}
		}
	}
	# Get Orphans
	foreach ($sf in $OriginalFiles) {
		if ($sf.MyFlag -eq "?") {
			$sf.MyFlag = "new"
			$FilesNew += $sf.Path
			Blabla "New`t$($sf.MyPath)`t$($sf.LastWriteTime)"
			}
	}
	foreach ($df in $CloudFiles) {
		if ($df.MyFlag -eq "?") {
			$df.MyFlag = "old"
			$FilesOld += $($df.Path + " ["+$df.MyPath+"]")
			Blabla "Old cloud`t$($df.MyPath)`t$($sf.LastWriteTime)`t$($df.Name)"
			}
	}
	foreach ($df in $LocalFiles) {
		if ($df.MyFlag -eq "?") {
			$df.MyFlag = "old"
			$FilesOld += $($df.Path + " ["+$df.MyPath+"]")
			Blabla "Old local`t$($df.MyPath)`t$($sf.LastWriteTime)"
			}
	}

	# Summary
	Blabla "Summary: identical $($FilesSame.count), changed $($FilesChanged.count), added $($FilesNew.count) and old $($FilesOld.count) files" ".*"

	# Talk
	#Blabla "Identical ones $($FilesSame.count)"
	#foreach ($f in $FilesSame)     { Blabla " $($f)"; }
	#Blabla "Changed ones $($FilesChanged.count)"
	#foreach ($f in $FilesChanged)  { Blabla " $($f)"; }
	#Blabla "New ones $($FilesNew.count)"
	#foreach ($f in $FilesNew)      { Blabla " $($f)"; }
	#Blabla "Old ones $($FilesOld.count)"
	#foreach ($f in $FilesOld)      { Blabla " $($f)"; }
	#Blabla "Original"
	#foreach ($f in $OriginalFiles) { Blabla " $($f.MyFlag) $($f.Path) $($f.MyPath)"; }
	#Blabla "Cloud"
	#foreach ($f in $CloudFiles)    { Blabla " $($f.MyFlag) $($f.Path) $($f.MyPath)"; }
	#Blabla "Local"
	#foreach ($f in $LocalFiles)    { Blabla " $($f.MyFlag) $($f.Path) $($f.MyPath)"; }
	
	Blabla "Updating $($FilesChanged.count) files..."
	foreach ($f in $FilesChanged)  { 
		$src = $f
		Blabla "Updating $($src)" "."
		BackupOne $src
		$cc = $cc +1
	}
	Blabla "Adding $($FilesNew.count) files..."
	foreach ($f in $FilesNew)  { 
		$src = $f
		Blabla "Adding $($src)" "."
		BackupOne $src
		$ca = $ca +1
	}
	Blabla "Removing $($FilesOld.count) files..."
	foreach ($f in $FilesOld)  { 
		mkdir $MyGone
		$src = $f
		$src = $src.substring(0,$src.IndexOf(" ["))
		$dst = $src.replace($MyCloud,$MyGone).replace($MyLocal,$MyGone)
		mkpath $dst
		Blabla "Removing $($src) $($dst)" "."
		move-item $src $dst
		$cd = $cd +1
	}
	Blabla "Backupped $($Mask) - changed $($cc), added $($ca) and removed $($cd) files" ".*"

	Cleanup $MyCloud
	Cleanup $MyLocal
}

# Update from backup (cloud or local) to cloud, USB drive or NAS
# Second param can be set to "?" to just look (and log all matches)
function Transfer([string] $Where,[string] $LookOnly) {
	$cc = 0; $ca = 0; $cd = 0;
	if ($LookOnly -ne "?") { Blabla "Transferring to $($Where)" "*"; }
	else                   { Blabla "Just checking $($Where)" "*";   }
	switch ($Where) {
	"Cloud" { $Source = $MyCloud;    $Destination = $ToCloud; break; }
	"USB"   { $Source = $MyLocal;    $Destination = $ToUSB;   break; }
	"NAS"   { $Source = $MyOriginal; $Destination = $ToNAS;   break; }
	default { exit; }
	}

	mkdir $Source
	mkdir $Destination

	# Get the files and add MyFlag and for the cloud ones the decoded name field
	Blabla "Scanning source from $($Source)"
	$SourceFiles      = Invoke-FastFind -Path $Source      -Recurse -File -System -Hidden -AttributeFilterMode Exclude | Select Name,LastWriteTime,Path,MyFlag,MyPath
	Blabla "Scanning desintation from $($Destination)"
	$DestinationFiles = Invoke-FastFind -Path $Destination -Recurse -File -System -Hidden -AttributeFilterMode Exclude | Select Name,LastWriteTime,Path,MyFlag,MyPath
	Blabla "... I found $($SourceFiles.count) source and $($DestinationFiles.count) destination files"
	foreach ($sf in $SourceFiles) {
		$sf.MyPath = $($sf.Path).replace($Source+"\","")
		$sf.MyFlag  = "?"
	}
	foreach ($df in $DestinationFiles) {
		$df.MyPath = $($df.Path).replace($Destination+"\","")
		$df.MyFlag  = "?"
	}
	$FilesSame = @()
	$FilesChanged = @()
	$FilesNew = @()
	$FilesOld = @()
	
	# Get the like named ones
	foreach ($sf in $SourceFiles) {
		foreach ($df in $DestinationFiles) {
			if ($sf.MyPath -eq $df.MyPath) {
				# A second deviation can happen (at least on my NAS) ... so we just check the source is not newer as the remote file
				$dif = $df.lastwritetime.subtract($sf.lastwritetime).totalseconds
				# Ignore up to 1 second difference
				if (($dif -le 1) -and ($dif -ge -1)) { $dif = 0 }
				$x = "{0:N0}" -f $dif
				# Since we really copy the files, timestamp and side are wanted identical
				if (($dif -eq 0) -and ($sf.Length -eq $df.Length)) {
					$sf.MyFlag = "same"
					$df.MyFlag = "same"
					$FilesSame += $sf.Path
					if ($LookOnly -eq "?") { 
						if ($Where -eq "Cloud") {
							Blabla "Same`t$($sf.Path)`t$($sf.LastWriteTime)`t$($df.LastWriteTime)`t$($x)`t$(Decode($sf.Name))"
						}
						else {
							Blabla "Same`t$($sf.Path)`t$($sf.LastWriteTime)`t$($df.LastWriteTime)`t$($x)" 
						}
					}
				}
				else {
					$sf.MyFlag = "changed"
					$df.MyFlag = "changed"
					$FilesChanged += $sf.Path
					if ($Where -eq "Cloud") {
						Blabla "Changed`t$($sf.Path)`t$($sf.LastWriteTime)`t$($df.LastWriteTime)`t$($x)`t$(Decode($sf.Name))"
					}
					else {
						Blabla "Changed`t$($sf.Path)`t$($sf.LastWriteTime)`t$($df.LastWriteTime)`t$($x)"
					}
				}
			}
		}
	}
	# Get Orphans
	foreach ($sf in $SourceFiles) {
		if ($sf.MyFlag -eq "?") {
			$sf.MyFlag = "new"
			$FilesNew += $sf.Path
			if ($Where -eq "Cloud") {
				Blabla "New`t$($sf.Path)`t$($sf.LastWriteTime)`t$(Decode($sf.Name))"
			}
			else {
				Blabla "New`t$($sf.Path)`t$($sf.LastWriteTime)"
			}
			}
	}
	foreach ($df in $DestinationFiles) {
		if ($df.MyFlag -eq "?") {
			$df.MyFlag = "old"
			$FilesOld += $df.Path
			if ($Where -eq "Cloud") {
				Blabla "Old`t$($df.Path)`t$($df.LastWriteTime)`t$(Decode($df.Name))"
			}
			else {
				Blabla "Old`t$($df.Path)`t$($df.LastWriteTime)"
			}
			}
	}

	# Summary
	Blabla "Summary: identical $($FilesSame.count), changed $($FilesChanged.count), added $($FilesNew.count) and old $($FilesOld.count) files" ".*"

	# Talk
	#Blabla "Identical ones $($FilesSame.count)"
	#foreach ($f in $FilesSame)     { Blabla " $($f)"; }
	#Blabla "Changed ones $($FilesChanged.count)"
	#foreach ($f in $FilesChanged)  { Blabla " $($f)"; }
	#Blabla "New ones $($FilesNew.count)"
	#foreach ($f in $FilesNew)      { Blabla " $($f)"; }
	#Blabla "Old ones $($FilesOld.count)"
	#Blabla "Source"
	#foreach ($f in $SourceFiles)      { Blabla " $($f.MyFlag) $($f.Path) $($f.MyPath)"; }
	#Blabla "Destination"
	#foreach ($f in $DestinationFiles) { Blabla " $($f.MyFlag) $($f.Path) $($f.MyPath)"; }

	if ($LookOnly -ne "?") {
		Blabla "Updating $($FilesChanged.count) files..."
		foreach ($f in $FilesChanged)  { 
			$src = $f
			$dst = $src.replace($Source,$Destination)
			mkpath $dst
			Blabla "Updating $($src)" "."
			Copy-Item $src $dst -Force
			$cc = $cc +1
		}
		Blabla "Adding $($FilesNew.count) files..."
		foreach ($f in $FilesNew)  { 
			$src = $f
			$dst = $src.replace($Source,$Destination)
			mkpath $dst
			Blabla "Adding $($src)" "."
			Copy-Item $src $dst -Force
			$ca = $ca +1
		}
		Blabla "Removing $($FilesOld.count) files..."
		foreach ($f in $FilesOld)  { 
			mkdir $MyGone
			$src = $f
			$dst = $src.replace($Destination,$MyGone+"\"+$Where)
			mkpath $dst
			if (Test-Path $dst) { Remove-Item $dst }
			Blabla "Removing $($src) $($dst)" "."
			move-item $src $dst
			$cd = $cd +1
		}
		Blabla "Transferred to $($Where) - changed $($cc), added $($ca) and removed $($cd) files" ".*"
		
		# In cloud we have A-Z as folders only, so why clean up there
		if ($Where -ne "Cloud") { Cleanup $Destination; } 
	}
}

function Outlook() {
	$c = 0
	Blabla "Getting Outlook files" "*"
	$dst = $MyOutlook + "\*"
	Remove-Item $dst 
	$files = (Get-ChildItem -Path $MyOST -File) + (Get-ChildItem -Path $MyPST -File)
	foreach ($f in $files) {
		$srcName = $f.Name
		$srcPath = $f.VersionInfo.FileName
		$dstPath = $MyOutlook + "\" + $srcName + ".zip"
		$wrkFolder = "-w" + (Split-Path $dstPath)
		#Command "C:\Program Files\7-Zip\7z.exe"
		# a					Add to archive
		# Archivname
		# -tzip				Type of archive
		# -w				Work folder
		# -y				Do not ask nothing
		# Sourcefile
		if (Test-Path $dstPath) { Remove-Item $dstPath }
		Blabla "Outlook adding`t$($srcPath)`t$($dstPath)" "."
		& 'C:\Program Files\7-Zip\7z.exe' a $dstPath -slp -tzip $wrkFolder -y $srcPath | Out-Null
		$c = $c +1
	}
	Blabla "Gotten Outlook files - $($c) files" ".*"
}

# Remove emtpy subfolders
function Cleanup([string]$What) {

	Blabla "Cleanup $($What)" "*"

	do {
		$empties = Invoke-FastFind -Path $What -Recurse -Directory -System -Hidden -AttributeFilterMode Exclude | select -ExpandProperty Path | where { (Get-ChildItem $_).count -eq 0 }
		$empties
		foreach ($d in $empties) { 
			Remove-Item $d
			Blabla "Dropped empty $($d)" "."
		}
	} until (!($empties))
}

function Cloud([string]$What) {
	switch ($What) {
	"on"  { net use $ToCloud.substring(0,2) /d > $null 2>&1
		    net use $ToCloud.substring(0,2) $MyCloudPath /user:$MyCloudUser $MyCloudPass
		    break; }
	"off" { net use $ToCloud.substring(0,2) /d  > $null 2>&1
		    break; }
	}
}

# Here we go
$MyLog = $MyLog.replace("*",$ParmCommand).replace("?","_");
$MyOld = $MyLog.replace(".log",".old")
if (Test-Path $MyOld) { Remove-Item $MyOld        | Out-Null }
if (Test-Path $MyLog) { Move-Item   $MyLog $MyOld | Out-Null }
Blabla "Started" "*"


switch ($ParmCommand) {
# The ones used regularly
"ba"	{ Outlook; Backup;        break; } # Snychronize original and outlook to backup folder
"bo"	{ Outlook;                break; } # Snychronize outlook to backup folder
"bh"	{ Backup;                 break; } # Snychronize original to backup folder
"tc"	{ Transfer "Cloud";       break; } # Snychronize local backup folder with cloud
"tu"	{ Transfer "USB";         break; } # Snychronize local backup folder with USB drive
"tn"	{ Transfer "NAS";         break; } # Snychronize local backup folder with NAS
"c+"	{ Cloud "on";             break; } # Mount cloud drive
"c-"	{ Cloud "off";            break; } # Unmount cloud drive

# The ones hopefully never needed
"lb"	{ ListBackup $ParmMask;   break; } # List files in original folder
"lo"	{ ListOriginal $ParmMask; break; } # List files in folder for backup to cloud, USB or NAS
"rs"	{ Restore $ParmMask;      break; } # Recover files from local backup

# The odd one just in case I want to just validate
"cc"	{ Transfer "Cloud" "?";   break; } # Simulate Snychronize local backup folder with cloud
"cu"	{ Transfer "USB" "?";     break; } # Simulate Snychronize local backup folder with USB drive
"cn"	{ Transfer "NAS" "?";     break; } # Simulate Snychronize local backup folder with NAS

# Hints
default { write-host "Try one of these:"
		  write-host "ba: Backup   Data & Outlook"
		  write-host "bh: Backup   Data"
		  write-host "bo: Backup   Outlook"
		  write-host "tc: Transfer Cloud"
		  write-host "tu: Transfer USB"
		  write-host "tn: Transfer NAS" 
		  write-host "cc: Check    Cloud"
		  write-host "cu: Check    USB"
		  write-host "cn: Check    NAS" 
		  if (Test-Path $MyLog) { Remove-Item -Path $MyLog | Out-Null }
		  break; }

}
Blabla "Done" "*"

